# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2010
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import ctypes
from ctypes import *
from . import findlib
from . import avutil
from . import decls
from .decls import AVCodec, AVFrame


class AVCodecError(avutil.AVError):
    pass


def avcodec_version():
    """Return the libavcodec library version.
    
    Returns a tuple (major,minor,micro) containing the three parts of the
    version number. 
    """
    v = _lib().avcodec_version()
    major = v>>16
    minor = (v>>8)&0xff
    micro = v&0xff
    return (major,minor,micro)

def avcodec_find_decoder(id):
    """Finds a decoder with a matching codec ID.
    
    id is an integer containing the codec id.
    Returns a AVCodec object or None if no decoder was found.
    """
    # The returned pointer is owned by the library, so we don't have to
    # do anything to get rid of it.
    func = _lib().avcodec_find_decoder
    func.restype = POINTER(AVCodec)
    res = func(id)
    if res:
        return res.contents
    else:
        return None

def avcodec_find_decoder_by_name(name):
    """Finds a registered decoder with the specified name.
    
    Returns a AVCodec object or None if no decoder was found.
    """
    # The returned pointer is owned by the library, so we don't have to
    # do anything to get rid of it.
    func = _lib().avcodec_find_decoder_by_name
    func.args = [ctypes.c_char_p]
    func.restype = POINTER(AVCodec)
    res = func(name)
    if res:
        return res.contents
    else:
        return None

def avcodec_find_encoder(id):
    """Finds a registered encoder with a matching codec ID.
    
    id is an integer containing the codec id.
    Returns an AVCodec object or None if no encoder was found.
    """
    # The returned pointer is owned by the library, so we don't have to
    # do anything to get rid of it.
    func = _lib().avcodec_find_encoder
    func.restype = POINTER(AVCodec)
    res = func(id)
    if res:
        return res.contents
    else:
        return None

def avcodec_find_encoder_by_name(name):
    """Finds a registered encoder with the specified name.
    
    Returns an AVCodec object or None if no encoder was found.
    """
    # The returned pointer is owned by the library, so we don't have to
    # do anything to get rid of it.
    func = _lib().avcodec_find_encoder_by_name
    func.args = [ctypes.c_char_p]
    func.restype = POINTER(AVCodec)
    res = func(name)
    if res:
        return res.contents
    else:
        return None

def avcodec_open(avctx, codec):
    """Initializes the AVCodecContext to use the given AVCodec. 

    avctx is a AVCodecContext object and codec a AVCodec object.
    Raises an exception when an error occurs.

    Prior to using this function the context has to be allocated.

    The functions avcodec_find_decoder_by_name(), avcodec_find_encoder_by_name(),
    avcodec_find_decoder() and avcodec_find_encoder() provide an easy way for 
    retrieving a codec.
    
    Warning: This function is not thread safe!
    """
    ret = _lib().avcodec_open(byref(avctx), byref(codec))
    if ret<0:
        # It can happen that the file was actually opened, so try to close it again.
        # The line is commented out because closing it here seems to cause more harm
        # (sometimes it leads to the lib not being able to open anything at all anymore)
        #avcodec_close(avctx)
        raise AVCodecError(ret)
    return ret    
    
def avcodec_close(codecCtx):
    """Close the codec.
    """
    return _lib().avcodec_close(byref(codecCtx))
    
def avcodec_alloc_frame():
    """Allocates an AVFrame and sets its fields to default values. 

    The resulting struct can be deallocated by simply calling avutil.av_free().
    Returns an AVFrame object.
    """
    func = _lib().avcodec_alloc_frame
    func.restype = POINTER(AVFrame)
    res = func()
    if res:
        return res.contents
    else:
        return None

def av_init_packet(pkt):
    """Initialize optional fields of a packet with default values.
    
    *pkt* is a :class:`AVPacket` object.
    Don't call this function on a packet that already contains data as this
    would lead to a memory leak.
    """
    _lib().av_init_packet(byref(pkt))
        
def av_free_packet(pkt):
    """Free a packet.
    
    *pkt* is a :class:`AVPacket` object.
    Frees the data associated with the packet (if there is any data set).
    """
    _lib().av_free_packet(byref(pkt))

#def avcodec_decode_video(codecCtx, picture, buf, bufsize):
#    """Decodes a video frame from buf into picture. 
#
#    The avcodec_decode_video() function decodes a video frame from the input
#    buffer buf of size buf_size. To decode it, it makes use of the video
#    codec which was coupled with avctx using avcodec_open(). The resulting 
#    decoded frame is stored in picture.    
#    Returns a tuple (got_picture, bytes) where got_picture is a boolean that
#    indicates whether a frame was decoded (True) or not and bytes is the number
#    of bytes used.
#    """
#    got_picture = c_int()
#    ret = _lib().avcodec_decode_video(byref(codecCtx), byref(picture), byref(got_picture), buf, bufsize)
#    if ret<0:
#        raise AVCodecError("Error: %s"%ret)
#    return bool(got_picture.value), ret

def avcodec_decode_video2(codecCtx, picture, pkt):
    """Decodes a video frame from pkt into picture. 

    Some decoders may support multiple frames in a single AVPacket, such
    decoders would then just decode the first frame.
    
    Some codecs have a delay between input and output, these need to be fed
    with avpkt.data=NULL, avpkt.size=0 at the end to return the remaining frames.

    *codecCtx* is an :class:`AVCodecCtx` object, *picture* is an :class:`AVFrame`
    object which will receive the decoded video frame. *pkt* is a :class:`AVPacket`
    object containing the encoded data.
    
    Returns a tuple (*got_picture*, *bytes*) where *got_picture* is a boolean that
    indicates whether a frame was decoded (True) or not and *bytes* is the number
    of bytes used or 0 if no frame could be decompressed.
    """
    got_picture = c_int()
    ret = _lib().avcodec_decode_video2(byref(codecCtx), byref(picture), byref(got_picture), byref(pkt))
    if ret<0:
        raise AVCodecError(ret)
    return bool(got_picture.value), ret

#def avcodec_decode_audio2(codecCtx, sampleBuf, buf, bufsize):
#    """Decode an audio frame.
#    
#    sampleBuf is a ctypes short array.
#    buf is the encoded data and bufsize the size of the encoded data.
#    Returns the frameSize (size in bytes of the decoded frame) and the number
#    of bytes that have been used from buf.
#    """
#    frameSize = c_int(ctypes.sizeof(sampleBuf))
#    ret = _lib().avcodec_decode_audio2(byref(codecCtx), sampleBuf, byref(frameSize), buf, bufsize)
#    if ret<0:
#        raise AVCodecError(ret)
#    return frameSize.value, ret

def avcodec_decode_audio3(codecCtx, sampleBuf, pkt):
    """Decode an audio frame.
    
    Some decoders may support multiple frames in a single AVPacket, such
    decoders would then just decode the first frame. In this case,
    avcodec_decode_audio3 has to be called again with an AVPacket that contains
    the remaining data in order to decode the second frame etc.
    
    *codecCtx* is an :class:`AVCodecCtx` object, *sampleBuf* is a ctypes short
    array that will receive the decoded audio data.
    *pkt* is a :class:`AVPacket` object containing the encoded data.
    
    Returns the frameSize (size in bytes of the decoded frame) and the number
    of bytes that have been used from the input buffer in *pkt*.
    If there are no more frames, both values are 0.
    """
    # The size of the output buffer in bytes
    frameSize = c_int(ctypes.sizeof(sampleBuf))
    ret = _lib().avcodec_decode_audio3(byref(codecCtx), sampleBuf, byref(frameSize), byref(pkt))
    if ret<0:
        raise AVCodecError(ret)
    return frameSize.value, ret

def avcodec_encode_video(codecCtx, buf, bufsize, picture):
    """Encodes a video frame from picture into buf.

    *codecCtx* is a :class:`AVCodecContext` object, *buf* is a pointer to
    the output buffer, *bufsize* is the size in bytes of the buffer and
    *picture* is a :class:`AVFrame` object containing the picture to encode.
    
    The input picture should be stored using a specific format, namely
    avctx.pix_fmt.
    Returns 0 or the number of bytes used from the output buffer.
    """
    ret = _lib().avcodec_encode_video(byref(codecCtx), buf, bufsize, byref(picture))
    if ret<0:
        raise AVCodecError(ret)
    return ret

def avpicture_alloc(picture, pix_fmt, width, height):
    """Allocate memory for a picture. 

    Call avpicture_free() to free it.
    
    picture     the AVPicture obejct to be filled in 
    pix_fmt     the format of the picture 
    width     the width of the picture 
    height     the height of the picture
    """
    ret = _lib().avpicture_alloc(byref(picture), pix_fmt, width, height)
    if ret<0:
        raise RuntimeError("Error: %s"%ret)
    
def avpicture_free(picture):
    """Free a picture previously allocated by avpicture_alloc().
    """
    _lib().avpicture_free(byref(picture))
    
def avpicture_get_size(pix_fmt, width, height):
    """Calculate image size.
    
    Calculate the size in bytes that a picture of the given 
    width and height would occupy if stored in the given picture format.
    
    pix_fmt     the given picture format 
    width     the width of the image 
    height     the height of the image
    
    Returns the size in bytes.
    """
    ret = _lib().avpicture_get_size(pix_fmt, width, height)
    if ret<0:
        raise AVCodecError(ret)
    return ret

def avpicture_fill(picture, ptr, pix_fmt, width, height):
    """Fill in the AVPicture fields. 

    The fields of the given AVPicture are filled in by using the 'ptr'
    address which points to the image data buffer. Depending on the
    specified picture format, one or multiple image data pointers and
    line sizes will be set. If a planar format is specified, several
    pointers will be set pointing to the different picture planes and
    the line sizes of the different planes will be stored in the
    lines_sizes array.
    
    picture     AVPicture whose fields are to be filled in 
    ptr     Buffer which will contain or contains the actual image data 
    pix_fmt     The format in which the picture data is stored. 
    width     the width of the image in pixels 
    height     the height of the image in pixels  

    Returns the size of the image data in bytes
    """
    return _lib().avpicture_fill(byref(picture), ptr, pix_fmt, width, height)

def avcodec_flush_buffers(codecCtx):
    """Flush buffers.

    Should be called when seeking or when switching to a different stream.
    """
    _lib().avcodec_flush_buffers(byref(codecCtx))

def avcodec_get_context_defaults2(codecCtx, codecType):
    """Set context defaults.
    
    In avcodec.h this function is marked as not being public, but what
    is the official way to do the same thing? (we are using it anyway
    as it's also used by the ffmpeg command line tool which we use as
    a reference)
    
    codecType is CODEC_TYPE_AUDIO, CODEC_TYPE_VIDEO, etc.
    """
    _lib().avcodec_get_context_defaults2(byref(codecCtx), codecType);

######################################################################

# By default, try to load the avcodec library that has the same major version
# than the one that was used for creating the cppdefs and decls module.
_libname = "avcodec.%s"%(decls.LIBAVCODEC_VERSION_MAJOR)
_libavcodec = None

def _lib():
    """Return the avcodec shared library.
    """
    global _libavcodec
    global _libname
    if _libavcodec is None:
        _libavcodec = findlib.findFfmpegLib(_libname)
    return _libavcodec




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 16 11:47:40 2019

@author: psakicki
"""

# _____  _____ _   _ ________   __  _____                      _                 _
#|  __ \|_   _| \ | |  ____\ \ / / |  __ \                    | |               | |
#| |__) | | | |  \| | |__   \ V /  | |  | | _____      ___ __ | | ___   __ _  __| | ___ _ __
#|  _  /  | | | . ` |  __|   > <   | |  | |/ _ \ \ /\ / / '_ \| |/ _ \ / _` |/ _` |/ _ \ '__|
#| | \ \ _| |_| |\  | |____ / . \  | |__| | (_) \ V  V /| | | | | (_) | (_| | (_| |  __/ |
#|_|  \_\_____|_| \_|______/_/ \_\ |_____/ \___/ \_/\_/ |_| |_|_|\___/ \__,_|\__,_|\___|_|






########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
from ftplib import FTP
import glob
import itertools
import multiprocessing as mp
import os 
import re
import shutil
import urllib

#### geodeZYX modules
from geodezyx import conv
from geodezyx import utils

#### Import star style
from geodezyx import *                   # Import the GeodeZYX modules
from geodezyx.externlib import *         # Import the external modules
from geodezyx.megalib.megalib import *   # Import the legacy modules names

##########  END IMPORT  ##########



############################################################################
######## RINEX DOWNLOADER
############################################################################


def igs_garner_server(stat,date):
    # plante si trop de requete
    urlserver = "ftp://garner.ucsd.edu/pub/rinex/"
    rnxname = conv.statname_dt2rinexname(stat.lower(),date)
    url = os.path.join( urlserver , str(date.year) , conv.dt2doy(date) , rnxname )
    return url

def igs_cddis_server(stat,date):
    # a privilegier
    urlserver = "ftp://cddis.gsfc.nasa.gov/gps/data/daily/"
    rnxname = conv.statname_dt2rinexname(stat.lower(),date)
    url = os.path.join(urlserver , str(date.year) , conv.dt2doy(date) , date.strftime('%y') + 'd' , rnxname)
    return url

def igs_cddis_nav_server(stat,date):
    # a privilegier
    urlserver = "ftp://cddis.gsfc.nasa.gov/gps/data/daily/"
    rnxname = conv.statname_dt2rinexname(stat.lower(),date,'n.Z')
    url = os.path.join(urlserver , str(date.year) , conv.dt2doy(date) , date.strftime('%y') + 'n' , rnxname)
    return url

def rgp_ign_smn_server(stat,date):
    urlserver = "ftp://rgpdata.ign.fr/pub/data/"
    rnxname = conv.statname_dt2rinexname(stat.lower(),date)
    url = os.path.join(urlserver , str(date.year) , conv.dt2doy(date) , 'data_30' , rnxname)
    return url

def rgp_ign_smn_1Hz_server(stat,date):
    urlserver = "ftp://rgpdata.ign.fr/pub/data/"

    urls = []

    for h in range(24):
        date_session = date
        date_session = date_session.replace(hour=h)

        print(date_session , 'session' , h)
        rnxname = conv.statname_dt2rinexname(stat.lower(),date_session ,
                                             session_a_instead_of_daily_session = 1)
        url = os.path.join(urlserver , str(date.year) , conv.dt2doy(date) ,
                           'data_1' , rnxname)

        urls.append(url)

    return urls

def unavco_server(stat,date):
    urlserver='ftp://data-out.unavco.org/pub/rinex'
    rnxname = conv.statname_dt2rinexname(stat.lower(),date)
    url = os.path.join(urlserver , 'obs', str(date.year) , conv.dt2doy(date) , rnxname)
    return url

def renag_server(stat,date):
    urlserver = "ftp://renag.unice.fr/data/"
    rnxname = conv.statname_dt2rinexname(stat.lower(),date)
    url = os.path.join(urlserver , str(date.year) , conv.dt2doy(date) , rnxname)
    return url

def orpheon_server(stat,date,user='',passwd=''):
    urlserver = "ftp://renag.unice.fr/"
    rnxname = conv.statname_dt2rinexname(stat.lower(),date)
    url = os.path.join(urlserver , str(date.year) , conv.dt2doy(date) , rnxname)
    return url,user,passwd

def ovsg_server(stat,date,user='',passwd=''):
    if dt.datetime(2009,1,1) <= date <= dt.datetime(2014,2,10):
        urlserver = "http://webobs.ovsg.univ-ag.fr/rawdata/GPS-GPSDATA.backtemp_20140210/"
    else:
        urlserver = "http://webobs.ovsg.univ-ag.fr/rawdata/GPS/GPSDATA/"
    rnxname = conv.statname_dt2rinexname(stat.lower(),date)
    url = os.path.join(urlserver , str(date.year) , conv.dt2doy(date) , 'rinex' , rnxname)
    return url,user,passwd

def geoaus_server(stat,date):
    """ Geosciences Australia
        ex : ftp://ftp.ga.gov.au/geodesy-outgoing/gnss/data/daily/2010/10063/ """
    urlserver = "ftp://ftp.ga.gov.au/geodesy-outgoing/gnss/data/daily/"
    rnxname = conv.statname_dt2rinexname(stat.lower(),date)
    url = os.path.join(urlserver , str(date.year) , date.strftime('%y') + conv.dt2doy(date) , rnxname)
    return url

def sonel_server(stat,date):
    """ex : ftp://ftp.sonel.org/gps/data/2015/001/ """
    urlserver = 'ftp://ftp.sonel.org/gps/data/'
    rnxname = conv.statname_dt2rinexname(stat.lower(),date)
    url = os.path.join(urlserver,str(date.year),conv.dt2doy(date),rnxname)
    return url

def effective_save_dir(parent_archive_dir,stat,date,archtype ='stat'):
    """
    INTERNAL_FUNCTION

    archtype =
        stat
        stat/year
        stat/year/doy
        year/doy
        year/stat
        week/dow
        OR only '/' for a dirty saving in the parent folder
        ... etc ... """
    if archtype == '/':
        return parent_archive_dir

    out_save_dir = parent_archive_dir
    fff = archtype.split('/')
    year = str(date.year)
    doy = conv.dt2doy(date)
    week, dow = conv.dt2gpstime(date)
    for f in fff:
        out_save_dir = os.path.join(out_save_dir,eval(f))
    return out_save_dir


############################################################################
######## PRODUCTS DOWNLOADER
############################################################################


def force_weekly_file_fct(force_weekly_file,sp3clk,day_in):
    if force_weekly_file == False:
        day = day_in

    elif type(force_weekly_file) is str or type(force_weekly_file) is int:
        print("INFO : The weekly file will be downloaded (DoW =",force_weekly_file,")")
        print("       Check force_weekly_file option if you don't want it")
        day = force_weekly_file

    elif sp3clk in ("erp","sum") and force_weekly_file == True:
        print("INFO : The weekly file (DoW = 7) will be downloaded for " + sp3clk.upper())
        day = 7
        
    return day




def orbclk_cddis_server(date,center='igs', sp3clk = 'sp3', repro=0, mgex=False,
                        longname = False, force_weekly_file=False):
    """
    longname is experimental and only for MGEX yet !! (180426)
    """
    urlserver='ftp://cddis.gsfc.nasa.gov/pub/gps/products/'
    if mgex:
        urlserver = urlserver + 'mgex/'
    if repro == 0:
        rep_fldr = ''
    else:
        rep_fldr = 'repro' + str(repro)
    if repro != 0:
        center     = list(center)
        center[-1] = str(repro)
        center = ''.join(center)
    if center in ("cod","cof","co2","cf2") and sp3clk == "sp3":
        print("INFO : CODE orbit extension changed to eph")
        sp3clk = "eph"

    # date definition
    week, day = conv.dt2gpstime(date)

    ## force_weekly_file handeling
    day = force_weekly_file_fct(force_weekly_file_fct,sp3clk,day)

    if not longname: # e.g. gbm19903.sp3.Z
        if not 'igu' in center:
            orbname = center + str(week).zfill(4)  + str(day) +'.'+ sp3clk +'.Z'
            url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)
        else:
            igusuffix = center[-2:]
            center    = center[:-2]
            orbname = center + str(week).zfill(4)  + str(day)  + '_' + igusuffix +'.'+ sp3clk + '.Z'
            url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)
    else: # e.g. COD0MGXFIN_20180580000_01D_05M_ORB.SP3.gz
        if len(center) == 3:
            center = center.upper() + "0"
        else:
            center = center.upper()

        datelong = date.strftime("%Y%j")

        if "SHA" in center:
            if sp3clk == "sp3":
                sp3clk_long = "_01D_15M_ORB.SP3"
            elif sp3clk == "clk":
                sp3clk_long = "_01D_05M_CLK.CLK"
            elif sp3clk == "erp":
                sp3clk_long = "_03D_12H_ERP.ERP"
            elif sp3clk == "bia":
                sp3clk_long = "_01D_01D_OSB.BIA"
            elif sp3clk == "snx":
                sp3clk_long = "_01D_000_SOL.SNX"

            orbname = center + "MGXRAP_" + datelong + "0000"  + sp3clk_long + ".gz"
        else:
            if sp3clk == "sp3":
                sp3clk_long = "_01D_15M_ORB.SP3"
            elif sp3clk == "clk":
                sp3clk_long = "_01D_30S_CLK.CLK"
            elif sp3clk == "erp":
                sp3clk_long = "_03D_12H_ERP.ERP"
            elif sp3clk == "bia":
                sp3clk_long = "_01D_01D_OSB.BIA"
            elif sp3clk == "snx":
                sp3clk_long = "_01D_000_SOL.SNX"

            orbname = center + "MGXFIN_" + datelong + "0000"  + sp3clk_long + ".gz"

        url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)

    return url



def orbclk_ign_server(date,center='igs', sp3clk = 'sp3', repro=0, mgex=False,
                        longname = False,force_weekly_file=False):
    """
    longname is experimental and only for MGEX yet !! (180426)
    Must be merged with orbclk_cddis_server to make a big equivalent fct (180523)
    """
    urlserver='ftp://igs.ign.fr/pub/igs/products/'
    if mgex:
        urlserver = urlserver + 'mgex/'
    if repro == 0:
        rep_fldr = ''
    else:
        rep_fldr = 'repro' + str(repro)
    if repro != 0:
        center     = list(center)
        center[-1] = str(repro)
        center = ''.join(center)
    week, day = conv.dt2gpstime(date)

    ## force_weekly_file handeling
    day = force_weekly_file_fct(force_weekly_file_fct,sp3clk,day)

    if not longname: # e.g. gbm19903.sp3.Z
        if not 'igu' in center:
            orbname = center + str(week).zfill(4)  + str(day) +'.'+ sp3clk +'.Z'
            url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)
        else:
            igusuffix = center[-2:]
            center    = center[:-2]
            orbname = center + str(week).zfill(4)  + str(day)  + '_' + igusuffix +'.'+ sp3clk + '.Z'
            url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)
    else: # e.g. COD0MGXFIN_20180580000_01D_05M_ORB.SP3.gz
        if len(center) == 3:
            center = center.upper() + "0"
        else:
            center = center.upper()

        datelong = date.strftime("%Y%j")

        if "SHA" in center:
            if sp3clk == "sp3":
                sp3clk_long = "_01D_15M_ORB.SP3"
            elif sp3clk == "clk":
                sp3clk_long = "_01D_05M_CLK.CLK"
            elif sp3clk == "erp":
                sp3clk_long = "_03D_12H_ERP.ERP"
            elif sp3clk == "bia":
                sp3clk_long = "_01D_01D_OSB.BIA"

            orbname = center + "MGXRAP_" + datelong + "0000"  + sp3clk_long + ".gz"
        else:
            if sp3clk == "sp3":
                sp3clk_long = "_01D_15M_ORB.SP3"
            elif sp3clk == "clk":
                sp3clk_long = "_01D_30S_CLK.CLK"
            elif sp3clk == "erp":
                sp3clk_long = "_03D_12H_ERP.ERP"
            elif sp3clk == "bia":
                sp3clk_long = "_01D_01D_OSB.BIA"

            orbname = center + "MGXFIN_" + datelong + "0000"  + sp3clk_long + ".gz"

        url = os.path.join(urlserver, str(week).zfill(4) ,rep_fldr,orbname)

    return url

def orbclk_igscb_server(date,center='gfz', sp3clk = 'sp3',repro=0):
    urlserver='ftp://igscb.jpl.nasa.gov/pub/product/'
    if repro == 0:
        rep_fldr = ''
    elif repro == 3:
        rep_fldr = 'repro' + str(repro)
    if repro != 0:
        center     = list(center)
        center[-1] = str(repro)
        center = ''.join(center)
    week, day = conv.dt2gpstime(date)
    if not 'igu' in center:
        orbname = center + str(week).zfill(4) + str(day) +'.' + sp3clk + '.Z'
        url = os.path.join(urlserver,str(week).zfill(4) ,rep_fldr,orbname)
    else:
        igusuffix = center[-2:]
        center    = center[:-2]
        orbname = center + str(week).zfill(4) + str(day)  + '_' + igusuffix +'.' + sp3clk + '.Z'
        url = os.path.join(urlserver,str(week).zfill(4) ,rep_fldr,orbname)
    return url

def orbclk_gfz_local_server(date,center='gfz', sp3clk='sp3',repro=0):
    if repro == 0:
        urlserver = '/dsk/igs_archive/IGS/SAVE_PROD_1d/'
    elif repro == 3:
        urlserver = '/dsk/repro3/ARCHIVE/IGS/SAVE_PROD_1d/'
    else:
        print("ERR : check the repro !!!")
        raise Exception

    week, day = conv.dt2gpstime(date)
    
    orbname = center + str(week).zfill(4) + str(day) +'.' + sp3clk + '.Z'
    url = os.path.join(urlserver,str(week).zfill(4),orbname)
    
    return url

def downloader(url,savedir,force = False,
               check_if_file_already_exists_uncompressed=True):
    """
    general function to download a file

    INTERNAL_FUNCTION
    """
#    print url
    if type(url) is tuple:
        need_auth = True
        username = url[1]
        password = url[2]
        url = url[0]
    else:
        need_auth = False

    url_print = str(url)

    rnxname = os.path.basename(url)

    Pot_compress_files_list = [os.path.join(savedir , rnxname)]

    if check_if_file_already_exists_uncompressed:
        Pot_compress_files_list.append(os.path.join(savedir , rnxname.replace(".gz","")))
        Pot_compress_files_list.append(os.path.join(savedir , rnxname.replace(".Z","")))
        Pot_compress_files_list = list(set(Pot_compress_files_list))

    for f in Pot_compress_files_list:
        if os.path.isfile(f) and (not force):
            print("INFO :", os.path.basename(f) , "already exists locally ;)")
            return None
        
    ##### LOCAL FILE (particular case for GFZ)
    if os.path.isfile(url):
        print("INFO : downloader : the is a local file, a simple copy will be used")
        print("       URL :",url)
        shutil.copy(url,savedir)
    
    ##### REMOTE FILE (General case)
    elif ("ftp" in url) or ("http" in url) :
        # managing a authentification
        if need_auth:
            if 'ftp://' in url: # FTP with Auth
                url = url.replace('ftp://','ftp://' + username + ':' + password + '@')
                opener = urllib.request.build_opener()
            else: # HTTP with Auth
                password_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
                top_level_url = url
                password_mgr.add_password(None, top_level_url, username, password)
                handler = urllib.request.HTTPBasicAuthHandler(password_mgr)
                # create "opener" (OpenerDirector instance)
                opener = urllib.request.build_opener(handler)
        else: # FTP or HTTP without Auth
            opener = urllib.request.build_opener()
    
        # use the opener to fetch a URL
        try:
            f = opener.open(url)
        except (urllib.error.HTTPError , urllib.error.URLError):
            print("WARN :",rnxname,"not found on server :(")
            print(url_print)
            return ""
        print("INFO :" , rnxname ," found on server :)")
        data = f.read()
        if not os.path.exists(savedir):
            os.makedirs(savedir)
        outpath = os.path.join(savedir , rnxname)
        with open(outpath, "wb") as code:
            code.write(data)
        return_str = outpath
    else:
        print("ERR : something goes wrong with the URL")
        print("     ", url)
        return_str = ""


    # effective downloading (old version)
#    try:
#        f = urllib2.urlopen(url)
#    except urllib2.HTTPError:
#        print rnxname," not found :("
#        print url
#        return None
#    print rnxname," found :)"
#    data = f.read()
#    if not os.path.exists(savedir):
#        os.makedirs(savedir)
#    with open(os.path.join(savedir , rnxname), "wb") as code:
#        code.write(data)
    return return_str

def start_end_date_easy(start_year,start_doy,end_year,end_doy):
    start = conv.doy2dt(start_year,start_doy)
    end   = conv.doy2dt(end_year,end_doy)
    return start , end



def effective_save_dir_orbit(parent_archive_dir,calc_center,date,
                             archtype ='year/doy/'):
    """
    INTERNAL_FUNCTION

    archtype =
        stat
        stat/year
        stat/year/doy
        year/doy
        year/stat
        week/dow
        wkwwww : use a GFZ's CF-ORB wk<wwww> naming
        OR only '/' for a dirty saving in the parent folder
        ... etc ...
    """
    if archtype == '/':
        return parent_archive_dir

    out_save_dir = parent_archive_dir
    fff = archtype.split('/')
    year = str(date.year)
    doy = conv.dt2doy(date)
    week, dow = conv.dt2gpstime(date)

    for f in fff:
        if "wkwwww" in f:
            f_evaluated = "wk" + str(week).zfill(4)
        else:
            f_evaluated = eval(f)
        out_save_dir = os.path.join(out_save_dir,str(f_evaluated))
    return out_save_dir


def downloader_wrap(intup):
    downloader(*intup)
    return None

def multi_downloader_rinex(statdico,archive_dir,startdate,enddate,
                           archtype ='stat',parallel_download=4,
                           sorted_mode=False,user='',passwd=''):
    """
    Parameters
    ----------
    statdico : dict
        a statdico is a dictionary associating Archives Centers to list of stations

        Exemple:
            >>> statdico['archive center 1'] = ['STA1','STA2','STA3', ...]
            >>> statdico['archive center 2'] = ['STA2','STA1','STA4', ...]

        the supported archive center are (july 2015):
            igs (cddis center)

            igs_garner (for the garner center, but not very reliable)

            rgp (St Mandé center)

            rgp_1Hz (all the 24 hourly rinex for the day will be downloaded)

            renag

            ovsg

            unavco

            sonel

            geoaus (Geosciences Australia)

            nav or brdc as archive center allows to download nav files (using 'BRDC' as station name) from the CDDIS server


    archtype : str
        string describing how the archive directory is structured, e.g :

            stat

            stat/year

            stat/year/doy

            year/doy

            year/stat

            week/dow/stat

            ... etc ...


    sorted_mode : bool
        if False:
            using the map multiprocess fct so the download order will
            be scrambled
        if True:
            using the apply multiprocess fct so the download order will be
            in the chrono. order
        The scrambled (False) is better, bc. it doesn't create zombies processes


    user & passwd : str
        user & password for a locked server


    Returns
    -------
    url_list : list of str
        list of URLs

    savedir_list : list of str
        list of downloaded products paths
    """

    curdate = startdate

    pool = mp.Pool(processes=parallel_download)

    urllist = []
    savedirlist = []


    print("generating the list of potential RINEXs ...")
    while curdate <= enddate:
        for netwk , statlis in list(statdico.items()):
            for stat in statlis:
                stat = stat.lower()
                mode1Hz = False

                if netwk == 'igs':
                    url = igs_cddis_server(stat,curdate)
                elif netwk == 'igs_garner':
                    url = igs_garner_server(stat,curdate)
                elif netwk == 'rgp':
                    url = rgp_ign_smn_server(stat,curdate)
                elif netwk == 'rgp_1Hz':
                    urls = rgp_ign_smn_1Hz_server(stat,curdate)
                    mode1Hz = True
                elif netwk == 'renag':
                    url = renag_server(stat,curdate)
                elif netwk == 'orpheon':
                    url = orpheon_server(stat,curdate,user,passwd)
                elif netwk == 'ovsg':
                    url = ovsg_server(stat,curdate,user,passwd)
                elif netwk == 'unavco':
                    url = unavco_server(stat,curdate)
                elif netwk == 'sonel':
                    url = sonel_server(stat,curdate)
                elif netwk == 'geoaus':
                    url = geoaus_server(stat,curdate)
                elif netwk in ('nav' , 'brdc'):
                    url = igs_cddis_nav_server(stat,curdate)
                else:
                    print('WARN : unkwn server dic in the dico, skip ...')
                    continue

                savedir = effective_save_dir(archive_dir,stat,curdate,archtype)

                if not mode1Hz:
                    urllist.append(url)
                    savedirlist.append(savedir)
                else:
                    urllist = urllist + urls
                    savedirlist = savedirlist + [savedir] * len(urls)

        curdate = curdate + dt.timedelta(days=1)

    #savedirlist = [x for (y,x) in sorted(zip(urllist,savedirlist))]
    #urllist     = sorted(urllist)
    
    urllist,savedirlist = utils.sort_binom_list(urllist,savedirlist)

    print(" ... done")
    print(len(urllist),"potential RINEXs")

    if sorted_mode:
        results = [pool.apply_async(downloader , args=(u,sd)) for u,sd in zip(urllist,savedirlist)]
    else:
        _ = pool.map(downloader_wrap,list(zip(urllist,savedirlist)))


    localfiles_lis = []
    skiped_url = 0
    for url , savedir in zip(urllist,savedirlist):
        try:
            localfile = os.path.join(savedir,os.path.basename(url))
            if os.path.isfile(localfile):
                localfiles_lis.append(localfile)
        except:
            # because of a weird error
            # i = p.rfind('/') + 1
            # AttributeError: 'tuple' object has no attribute 'rfind'
            skiped_url += 1
            continue

    pool.close()
    print('INFO : ' , skiped_url , 'returned url skiped because of a weird error, but it is not important :)')
    return localfiles_lis

#    return zip(urllist,savedirlist)


def orbclk_long2short_name(longname_filepath_in,
                           rm_longname_file=False,
                           center_id_last_letter=None,
                           center_manual_short_name=None,
                           force=False,
                           dryrun=False,
                           output_dirname=None):
    """
    Rename a long naming new convention IGS product file to the short old
    convention
    Naming will be done automaticcaly based on the 3 first charaters of the
    long AC id
    e.g. CODE => cod, GRGS => grg, NOAA => noa ...

    Parameters
    ----------
    longname_filepath_in : str
        Full path of the long name product file

    rm_longname_file : bool
        Remove the original long name product file

    center_id_last_letter : str
        replace the last letter of the short AC id by another letter
        (see note below)
        
    center_manual_short_name : str
        replace completely the long name with this one
        overrides center_id_last_letter
        
    force : bool
        if False, skip if the file already exsists

    dryrun : bool
        if True, don't rename effectively, just output the new name
        
    output_dirname : str
        directory where the output shortname will be created
        if None, will be created in the same folder as the input longname

    Returns
    -------
    shortname_filepath : str
        Path of the  short old-named product file

    Note
    ----
    if you rename MGEX orbits, we advise to set
    center_id_last_letter="m"
    the AC code name will be changed to keep a MGEX convention
    (but any other caracter can be used too)

    e.g. for Bern's products, the long id is CODE

    if center_id_last_letter=None, it will become cod,
    if center_id_last_letter=m, it will become com

    """

    print("INFO : will rename" , longname_filepath_in)

    longname_basename = os.path.basename(longname_filepath_in)
    longname_dirname  = os.path.dirname(longname_filepath_in)
    
    if not output_dirname:
        output_dirname = longname_dirname

    center = longname_basename[:3]

    if center_manual_short_name:
        center = center_manual_short_name
    elif center_id_last_letter:
        center_as_list = list(center)
        center_as_list[-1] = center_id_last_letter
        center = "".join(center_as_list)

    yyyy   = int(longname_basename.split("_")[1][:4])
    doy    = int(longname_basename.split("_")[1][4:7])

    day_dt = conv.doy2dt(yyyy,doy)

    wwww , dow = conv.dt2gpstime(day_dt)

    shortname_prefix = center.lower() + str(wwww) + str(dow)

    ### Type handeling
    if   "SP3" in longname_basename:
        shortname = shortname_prefix + ".sp3"
    elif "CLK" in longname_basename:
        shortname = shortname_prefix + ".clk"
    elif "ERP" in longname_basename:
        shortname = shortname_prefix + ".erp"
    elif "BIA" in longname_basename:
        shortname = shortname_prefix + ".bia"
    elif "SNX" in longname_basename:
        shortname = shortname_prefix + ".snx"
    else:
        print("ERR : filetype not found for",longname_basename)
    
    ### Compression handeling
    if longname_basename[-3:] == ".gz":
        shortname = shortname + ".gz"
    elif longname_basename[-2:] == ".Z":
        shortname = shortname + ".Z"
        
    shortname_filepath = os.path.join(output_dirname , shortname)
    
    if not force and os.path.isfile(shortname_filepath):
        print("INFO : skip", longname_filepath_in)
        print("     ",shortname_filepath,"already exists")        
        return shortname_filepath
    
    if not dryrun:
        print("INFO : renaming" , longname_filepath_in,"=>",shortname_filepath)
        shutil.copy2(longname_filepath_in , shortname_filepath)

    if rm_longname_file and not dryrun:
        print("INFO : remove " , longname_filepath_in)
        os.remove(longname_filepath_in)

    return shortname_filepath



def rnx_long2short_name(longname_filepath_in):
    """
    MUST BE IMPROVED
    """
    
    longname_basename = os.path.basename(longname_filepath_in)
    longname_dirname  = os.path.dirname(longname_filepath_in)
    
    Longname_basename_splitted = longname_basename.split("_")
    
    datepart_str = Longname_basename_splitted[2]
    yyyy = datepart_str[:4]
    ddd  = datepart_str[4:7]

    shortname_basename = longname_basename[:4].lower() + ddd + "0." + yyyy[2:] + "o"
    
    return os.path.join(longname_dirname,shortname_basename)
    
    
    


def multi_downloader_orbs_clks(archive_dir,startdate,enddate,calc_center='igs',
                            sp3clk='sp3',archtype ='year/doy',parallel_download=4,
                            archive_center='cddis',repro=0,sorted_mode=False,
                            force_weekly_file=False, return_also_uncompressed_files=True):

    """
    Download IGS products. Can manage MGEX products too
    (see archive_center argument)

    Parameters
    ----------
    archive_dir : str
        Parent archive directory where files will be stored

    startdate & enddate : datetime
        Start and End of the wished period

    calc_center : str or list of str
        calc_center can be a string or a list, describing the calc center
        e.g. 'igs','grg','cod','jpl' ...

    sp3clk : str
        Product type, can handle :

            'clk'

            'clk_30s'

            'sp3'
            
            'snx'
            
            'sum'

            'erp'

            'bia'

    archive_center : str
        server of download, "regular" IGS or MGEX, can handle :

            'cddis'

            'cddis_mgex'

            'cddis_mgex_longname'

            'ign_mgex'

            'ign_mgex_longname'
            
            'gfz_local'

    archtype: str
        string describing how the archive directory is structured, e.g :

            stat

            stat/year

            stat/year/doy

            year/doy

            year/stat

            week/dow/stat

            ... etc ...

    repro : int
        number of the IGS reprocessing (0 = routine processing)

    sorted_mode : bool
        if False:
            using the map multiprocess fct so the download order will
            be scrambled
        if True:
            using the apply multiprocess fct so the download order will be
            in the chrono. order
        The scrambled (False) is better, bc. it doesn't create zombies processes

    Returns
    -------
    localfiles_lis : list of str
        list of downloaded products paths
    """

    if type(calc_center) is str:
        calc_center =  [ ''.join([calc_center]) ] # POURQUOI CETTE LIGNE ?? (150717)
#        calc_center =  [calc_center]

    pool = mp.Pool(processes=parallel_download)
    urllist = []
    savedirlist = []

    if sp3clk == 'clk':
        typ = 'Clocks'
    elif sp3clk == 'clk_30s':
        typ = 'Clocks (30s)'
    elif sp3clk == 'sp3':
        typ = 'Orbits'
    elif sp3clk == 'erp':
        typ = 'ERP'
    elif sp3clk == 'snx':
        typ = 'SINEXs'
    elif sp3clk == 'sum':
        typ = 'SUM files'
    elif sp3clk == 'bia':
        typ = 'ISBs'
    else:
        typ = '????'

    print("generating the list of potential " + typ + " ...")
    
    for cc in calc_center:
            curdate = startdate
            while curdate <= enddate:
                if re.search("igs([0-9][0-9]|yy|YY)P",cc):
                    cc = "igs" + str(curdate.year)[2:] + "P"
                    print("INFO : IGS reference frame snx/ssc, correcting the year : ",cc)

                url = ''
                if archive_center == 'cddis':
                    url = orbclk_cddis_server(curdate,cc,repro=repro,sp3clk=sp3clk,
                                              force_weekly_file=force_weekly_file)
                elif archive_center == 'cddis_mgex':
                    url = orbclk_cddis_server(curdate,cc,repro=repro,sp3clk=sp3clk,
                                              mgex=True,force_weekly_file=force_weekly_file)
                elif archive_center == 'cddis_mgex_longname':
                    url = orbclk_cddis_server(curdate,cc,repro=repro,
                                              sp3clk=sp3clk,mgex=True,longname=True,
                                              force_weekly_file=force_weekly_file)
                elif archive_center == 'ign_mgex':
                    url = orbclk_ign_server(curdate,cc,repro=repro,sp3clk=sp3clk,
                                              mgex=True,force_weekly_file=force_weekly_file)
                elif archive_center == 'ign_mgex_longname':
                    url = orbclk_ign_server(curdate,cc,repro=repro,
                                              sp3clk=sp3clk,mgex=True,longname=True,
                                              force_weekly_file=force_weekly_file)
                elif archive_center == 'gfz_local':
                    url = orbclk_gfz_local_server(curdate,cc,repro=repro,
                                              sp3clk=sp3clk)

                else:
                    print('ERR : Wrong archive_center name !!! :' , archive_center)
                urllist.append(url)
                savedir = effective_save_dir_orbit(archive_dir,cc,curdate,archtype)
                savedirlist.append(savedir)
                curdate = curdate + dt.timedelta(days=1)

    savedirlist = [x for (y,x) in sorted(zip(urllist,savedirlist))]
    urllist = sorted(urllist)

    print(" ... done")
    print(len(urllist),"potential " + typ)

#    if sorted_mode:
#        print 'aaaa'
#        results = [pool.apply_async(downloader , args=(u,sd)) for u,sd in zip(urllist,savedirlist)]
#    else:
#        print 'bb'
#        _ = pool.map(downloader_wrap,zip(urllist,savedirlist))
#
#
    if not sorted_mode:
        _ = pool.map(downloader_wrap,list(zip(urllist,savedirlist)))
    else:
         results = [pool.apply_async(downloader , args=(u,sd)) for u,sd in zip(urllist,savedirlist)]

    localfiles_lis = []
    
    if not return_also_uncompressed_files:
        for url , savedir in zip(urllist,savedirlist):
            localfile = os.path.join(savedir,os.path.basename(url))
            if os.path.isfile(localfile):
                localfiles_lis.append(localfile)
    else:
    
        for url , savedir in zip(urllist,savedirlist):

            localfile = os.path.join(savedir,os.path.basename(url))        
            
            Pot_compress_files_list = [localfile]
            Pot_compress_files_list.append(localfile.replace(".gz",""))
            Pot_compress_files_list.append(localfile.replace(".Z",""))
            Pot_compress_files_list = list(set(Pot_compress_files_list))
            
            for potential_exisiting_file in Pot_compress_files_list:
                if os.path.isfile(potential_exisiting_file):
                    localfiles_lis.append(potential_exisiting_file)
    
    pool.close()
    return localfiles_lis

def multi_finder_rinex(main_dir,rinex_types=('o','d','d.Z','d.z'),
                       specific_stats = [] ):
    """
    from a main_dir, find all the rinexs in this folder and his subfolder

    (corresponding to the rinex_types)

    and return a list of the found rinexs

    is very similar with geodetik.rinex_lister and  gins_runner.get_rinex_list

    But this one is the most elaborated , must be used in priority !!!
    """
    files_raw_lis , _ = utils.walk_dir(main_dir)

    yylis = [str(e).zfill(2) for e in list(range(80,100)) + list(range(0,dt.datetime.now().year - 2000 + 1))]

    rinex_lis = []


    for f in files_raw_lis:
        for rnxext in rinex_types:
            for yy in yylis:
                if f.endswith(yy + rnxext):
                    rinex_lis.append(f)


    # CASE FOR specific_stats
    if len(specific_stats) > 0:
        rinex_lis2 = []
        for stat in specific_stats:
            for rnx in rinex_lis:
                if stat in os.path.basename(rnx):
                    rinex_lis2.append(rnx)
        rinex_lis = rinex_lis2

    print('INFO : ' , len(rinex_lis) , 'RINEXs found')

    return rinex_lis


def multi_archiver_rinex(rinex_lis,parent_archive_dir,archtype='stat',
                         move=True,force_mv_or_cp=True):
    """
    from rinex_lis, a list of rinex (generated by the function
    multi_finder_rinex)

    move (if move=True) of copy (if move=False) those rinexs in the
    parent_archive_dir according to the archtype,
    string describing how the archive directory is structured, e.g :
            stat

            stat/year

            stat/year/doy

            year/doy

            year/stat

            week/dow/stat

            ... etc ...
    """

    mv_cnt   = 0
    skip_cnt = 0
    print('INFO : RINEXs as input :' , len(rinex_lis))

    if move:
        mv_fct = utils.move
    else:
        mv_fct = utils.copy2

    for rnx in rinex_lis:
        date = conv.rinexname2dt(rnx)
        rnxname = os.path.basename(rnx)
        stat = rnxname[0:4]
        savedir = effective_save_dir(parent_archive_dir, stat, date, archtype)
        utils.create_dir(savedir)

        if not force_mv_or_cp and os.path.isfile(os.path.join(savedir,rnxname)):
            skip_cnt += 1
            continue
        else:
            mv_fct(rnx,savedir)
            mv_cnt += 1

    print('INFO : RINEXs skiped   :' , skip_cnt, '(because already existing)')
    print('INFO : RINEXs moved    :' , mv_cnt)

    return None


def find_IGS_products_files(parent_dir,File_type,ACs,date_start,date_end=None,
                            recursive_search=True,severe=True,
                            compressed="incl",
                            regex_old_naming = True,
                            regex_new_naming = True,
                            regex_igs_tfcc_naming = True,
                            add_weekly_file = False):
    """
    Find all product files in a parent folder which correspond to file type(s),
    AC(s) and date(s)

    Parameters
    ----------
    parent_dir : str or list of str
        The parent directory (i.e. the archive) where files are stored
        can be a string (path of the archive) or a list of file paths
        (given by the function utils.find_recursive) in order to gain time


    File_type : str or list of str
        File type(s) researched (sp3, erp, clk ...)
        can be a list of string for several file type paths
        or a string like 'sp3' if only one file type is researched

    ACs : 'all' or str or list of str
        AC(s) researched
        can be a list of string for several ACs
        or a string like 'gfz' if only one AC is researched
        if 'all', search for all the ACs

    date_start : dt.datetime or 2-tuple list of int
        begining of the time period researched
        can be a datetime
        or a 2-tuple (wwww,d) e.g. (1990,0)

    date_end : None or dt.datetime or 2-tuple list of int
        end of the time period researched
        can be a datetime
        or a 2-tuple (wwww,d) e.g. (1990,0)
        if None, then only date_start is researched
    
    severe : bool
        If True, raises an exception if something goes wrong

    compressed : str
        How the compressed files are handled
        "incl": include the compressed files
        "only": only consider the compressed files
        "excl": exclude the compressed files
        
    regex_old_naming : bool
        Handle old naming format

    regex_new_naming : bool
        Handle new naming format        
    
    regex_igs_tfcc_naming : bool
        Handle TFCC specific format (for SINEX files)
        
    add_weekly_file : bool
        Also handle the weekly file (day 7)
        Implemented only for the  old naming format (for the moment)

    Returns
    -------
    Files_select_cumul_list : list
        List of files found

    """
    
    ###### Prelim Checks   ##############
    if not os.path.exists(parent_dir):
        print("ERR : parent directory doesn't exist")
        print(parent_dir)
        if severe:
            raise Exception

    ###### Time management ##############
    ###### work internally with datetime
    ### date_start
    if type(date_start) is dt.datetime:
        date_start_ok = date_start
    else:
        date_start_ok = conv.gpstime2dt(*date_start)
    ### date_end
    if not date_end:
        date_end_ok = date_start_ok
    elif type(date_end) is dt.datetime:
        date_end_ok = date_end
    else:
        date_end_ok = conv.gpstime2dt(*date_end)
    ### generate time period with a while loop
    Dates_list = [date_start_ok]
    while Dates_list[-1] < date_end_ok:
        Dates_list.append(Dates_list[-1]  + dt.timedelta(days=1))
        
    ### manage weekly file 
    Dates_wwwwd_list   = [utils.join_improved("",*conv.dt2gpstime(d)) for d in Dates_list]
    Dates_yyyyddd_list = [utils.join_improved("",*reversed(conv.dt2doy_year(d))) for d in Dates_list]
    
    

    ###### File type / ACs management ##############

    if not utils.is_iterable(File_type):
        File_type = [File_type]
    if not utils.is_iterable(ACs):
        ACs = [ACs]

    ###### General file search management ##############
    if utils.is_iterable(parent_dir):
        FILE_LIST = parent_dir
    elif recursive_search:
        # All the files are listed first
        FILE_LIST = utils.find_recursive(parent_dir,".*",case_sensitive=False)
    else:
        FILE_LIST = glob.glob(parent_dir + "/*")


    ###### Regex Definition ##############
    
    join_regex_and = lambda  L : "(" +  "|".join(L) + ")"
    
    Re_patt_big_stk = []
    
    
    ### compression handeling
    if compressed == "excl":
        re_patt_comp = "$"
    elif compressed == "incl":
        re_patt_comp = "(\.Z|\.gz|)$"
    elif compressed == "only":
        re_patt_comp = "(\.Z|\.gz)$"
    else:
        print("ERR : check 'compressed' keyword (excl,incl, or only)")
        raise Exception
        
    if regex_old_naming: 
        if ACs[0] == "all":
            re_patt_ac = "\w{3}"
        else:
            re_patt_ac = join_regex_and([ac.lower() for ac in ACs])
                 
        if add_weekly_file:
            Dates_wwwwd_list_4old = [ e[:-1] + "(" + e[-1] + "|" + "7)"  for e in Dates_wwwwd_list]
        else:
            Dates_wwwwd_list_4old = Dates_wwwwd_list
        
        re_patt_date   = join_regex_and(Dates_wwwwd_list_4old)
        re_patt_filtyp = join_regex_and(File_type)
        re_patt_big_old_naming = re_patt_ac + re_patt_date + "\." + re_patt_filtyp + re_patt_comp
        Re_patt_big_stk.append(re_patt_big_old_naming)
        
    if regex_new_naming: ### search for new name convention
        if ACs[0] == "all":
            re_patt_ac = "\W{3}"        
        else:
            re_patt_ac = join_regex_and([ac.upper() for ac in ACs])
        re_patt_date   = join_regex_and(["_"+e for e in Dates_yyyyddd_list]) #add _ because it can raise a conflit with the old format
        re_patt_filtyp = join_regex_and([fil.upper() for fil in File_type])
        re_patt_big_new_naming = ".*".join((re_patt_ac,re_patt_date,re_patt_filtyp + re_patt_comp))
        Re_patt_big_stk.append(re_patt_big_new_naming)
        
    if regex_igs_tfcc_naming:
        Dates_yy_list = list(set([str(conv.gpstime2dt(int(e[0:4]),int(e[4])).year)[2:] for e in Dates_wwwwd_list]))
        Dates_wwww_list = list(set([e[:-1] for e in Dates_wwwwd_list]))
        #Dates_wwww_dot_list = [e + "\." for e in Dates_wwww_list]
        re_patt_year = join_regex_and(Dates_yy_list)
        # 2x re_patt_date : because .sum doesn't the day
        re_patt_date = join_regex_and(Dates_wwwwd_list + Dates_wwww_list)
        re_patt_filtyp = "\." +  join_regex_and(File_type)

        re_patt_big_igs_tfcc_naming = "igs" + re_patt_year + "P" + re_patt_date + ".*" + re_patt_filtyp + re_patt_comp
        Re_patt_big_stk.append(re_patt_big_igs_tfcc_naming)

    re_patt_big = join_regex_and(Re_patt_big_stk)
        
    print("INFO : REGEX researched :")    
    print(re_patt_big)
    
    ###### Specific file search management ##############
    Files_select_list = []
    for fil in FILE_LIST:
        if re.search(re_patt_big,os.path.basename(fil),re.IGNORECASE):
            Files_select_list.append(fil)
            
    if len(Files_select_list) == 0:
        print("ERR : no products found")
        if severe:
            raise Exception

    print("INFO : number of files found :",len(Files_select_list))    
    print(re_patt_big)
    
    return Files_select_list

def FTP_downloader(ftp_obj,filename,localdir):   
    localpath = os.path.join(localdir,filename)
    
    if not utils.empty_file_check(localpath):
        print("INFO:",filename,"already exists ;)")
        bool_dl = True
    else:
        try:
            localfile = open(localpath, 'wb')
            ftp_obj.retrbinary('RETR ' + filename, localfile.write, 1024)
            localfile.close()
            bool_dl = True
            print("INFO:",filename,"downloaded :)")
    
        except:
            print("WARN:",localpath,"download failed :(")
            bool_dl = False
    
    return localpath , bool_dl

def FTP_downloader_wo_objects(tupin):
    arch_center_main,wwww_dir,filename,localdir = tupin
    ftp_obj_wk = FTP(arch_center_main)
    ftp_obj_wk.login()
    ftp_obj_wk.cwd(wwww_dir)
    localpath , bool_dl = FTP_downloader(ftp_obj_wk,filename,localdir)
    ftp_obj_wk.close()
    return localpath , bool_dl
    

def multi_downloader_orbs_clks_2(archive_dir,startdate,enddate,
                            AC_names = ("wum","cod"),
                            prod_types = ("sp3","clk"),
                            remove_patterns=("ULA",),
                            archtype ='week',
                            new_name_conv = True,
                            parallel_download=4,
                            archive_center='ign',
                            mgex=True,repro=0,sorted_mode=False,
                            return_also_uncompressed_files=True,
                            ftp_download=False,
                            dow_manu=False):
    """
    dow_manu = False, no dow manu, consider the converted dow from the time span, regular case
    dow_manu = None, no dow in the REGEX, the crawler will search only for the week
    dow_manu = 0 or 7: the dow in question    
    """
    
    if mgex:
        mgex_str = "mgex/"
    else:
        mgex_str = ""
        
    if not utils.is_iterable(remove_patterns):
        remove_patterns = [remove_patterns]
    
    if archive_center == "cddis":
        arch_center_main    = 'cddis.gsfc.nasa.gov'
        arch_center_basedir = '/pub/gps/products/' + mgex_str
        
    elif archive_center == "cddis_glonass":
        arch_center_main    = 'cddis.gsfc.nasa.gov'
        arch_center_basedir = '/pub/glonass/products/' + mgex_str
    
    elif archive_center == "ign":
        arch_center_main    = 'igs.ign.fr'
        arch_center_basedir = '/pub/igs/products/' + mgex_str  

    elif archive_center == "ensg":
        arch_center_main    = 'igs.ensg.ign.fr'
        arch_center_basedir = '/pub/igs/products/' + mgex_str
        
    elif archive_center == "whu":
        arch_center_main    = "igs.gnsswhu.cn"
        arch_center_basedir = "/pub/gps/products/" + mgex_str
        
    elif archive_center == "ign_rf":
        arch_center_main    = 'igs-rf.ign.fr'
        arch_center_basedir = '/pub/' + mgex_str  

    elif archive_center == "ensg_rf":
        arch_center_main    = 'igs-rf.ensg.ign.fr'
        arch_center_basedir = '/pub/' + mgex_str

    print("INFO : data center used :",archive_center)

    Dates_list = conv.dt_range(startdate,enddate)

    Localfiles_lis = []
    wwww_dir_previous = None
    pool = mp.Pool(processes=parallel_download) 
   

    ## create the FTP object
    ftp = FTP(arch_center_main)
    ftp.login()
    ## create a list of FTP object for multiple downloads (unstable...)
    if ftp_download and parallel_download > 1:
        Ftp_obj_list = [FTP(arch_center_main) for i in range(parallel_download)]
        [f.login() for f in Ftp_obj_list]    
        # define the main obj for crawling
        ftp = Ftp_obj_list[0]
    
    ### check if the pattern of the wished products are in the listed daily files
    for patt_tup in list(itertools.product(Dates_list,AC_names,prod_types)):
        dt_cur , ac_cur , prod_cur = patt_tup
        wwww , dow = conv.dt2gpstime(dt_cur)
        
        #### Manage the cases of manual DOW
        if type(dow_manu) is int:
            dow = dow_manu
        elif dow_manu is None:
            dow = ""
        elif dow_manu is False:
            pass        
        else:
            dow = str(dow_manu)
               
        print("INFO : ","Search products for day",wwww,dow,"AC/prod",ac_cur,prod_cur)
        wwww_dir = os.path.join(arch_center_basedir,str(wwww))
        print("       Move to:",wwww_dir)
        if wwww_dir_previous != wwww_dir:
            try:
                ftp.cwd(wwww_dir)
            except:
                print("WARN:",wwww_dir,"do not exists, skiping...")
            Files_listed_in_FTP = ftp.nlst()
            wwww_dir_previous = wwww_dir
            if len(Files_listed_in_FTP) == 0:
                print("WARN: no files found in directory",wwww_dir)
                
                
        Files_remote_date_list = []

        pattern_old_nam = ac_cur+".*"+str(wwww)+str(dow)+".*"+prod_cur+"\..*"
        Files = [f for f in Files_listed_in_FTP if re.search(pattern_old_nam,f)]
        
        pattern_new_nam = ""
        
        Files_new_nam = []
        if new_name_conv: ### search for new name convention
            
            if dow is None:
                print("ERR: dow == None and search for new name convention, Error ...")
                raise  Exception()
                
            ac_newnam   = ac_cur.upper()

            doy_newnam  = "".join(reversed(conv.dt2doy_year(conv.gpstime2dt(wwww,dow))))
            prod_newnam = prod_cur.upper()
            
            pattern_new_nam = utils.join_improved(".*",ac_newnam,doy_newnam,prod_newnam)
            pattern_new_nam = ".*" + pattern_new_nam + "\..*"

            Files_new_nam   = [f for f in Files_listed_in_FTP if re.search(pattern_new_nam,f)]
        
        
        print("      ","Regex :",pattern_old_nam,pattern_new_nam)        
        Files = Files + Files_new_nam
        
        if len(Files) == 0:
            print("WARN : ","no product found for",*patt_tup)
        
        Files_remote_date_list = Files_remote_date_list + Files
            
        ### exclude some pattern
        for negpatt in remove_patterns:
            Files_remote_date_list = [e for e in Files_remote_date_list if not re.search(negpatt,e)]

        archive_dir_specif = effective_save_dir_orbit(archive_dir,
                                                      ac_cur,
                                                      dt_cur,
                                                      archtype)
            
        utils.create_dir(archive_dir_specif)
        
        ### Generation of the Download fct inputs
        Files_remote_date_chunck = utils.chunkIt(Files_remote_date_list,
                                                 parallel_download)
        Downld_tuples_list = []
        Potential_localfiles_list = []

        if ftp_download: ### FTP download is not recommended
            for ftpobj , Chunk in zip(Ftp_obj_list,Files_remote_date_chunck):
                for filchunk in Chunk:
                        Potential_localfiles_list.append(os.path.join(archive_dir_specif,filchunk))
                        if parallel_download == 1:
                            Downld_tuples_list.append((ftpobj,filchunk,archive_dir_specif))
                        else:
                            Downld_tuples_list.append((arch_center_main,wwww_dir,
                                                       filchunk,archive_dir_specif))
        else: ### HTML download, recommended
            Downld_tuples_list = itertools.product(["/".join(('ftp://' + arch_center_main,wwww_dir,f)) for f in Files_remote_date_list],[archive_dir_specif])
            [Potential_localfiles_list.append(os.path.join(archive_dir_specif,f)) for f in Files_remote_date_list]

            
        ### Actual Download, FTP download is not recommended
        if ftp_download and parallel_download == 1:
            for tup in Downld_tuples_list:
                FTP_downloader(*tup)
        elif ftp_download and parallel_download > 1:
            _ = pool.map_async(FTP_downloader_wo_objects,Downld_tuples_list)
        elif not ftp_download and parallel_download == 1:
            for tup in Downld_tuples_list:
                downloader_wrap(tup)
        elif not ftp_download and parallel_download > 1:
            _ = pool.map(downloader_wrap,Downld_tuples_list)
        ## Those 2 other methods are unstables
        ##  _ = pool.map_async(downloader_wrap,Downld_tuples_list)
        ##  _ = [pool.apply(downloader_wrap,(tup,)) for tup in Downld_tuples_list]
    

    ### Independent files existence check
    if not return_also_uncompressed_files:
        for localfile in Potential_localfiles_list:
            if os.path.isfile(localfile):
                Localfiles_lis.append(localfile)
    else:
        for localfile in Potential_localfiles_list:
            Pot_compress_name_list = [localfile]
            Pot_compress_name_list.append(localfile.replace(".gz",""))
            Pot_compress_name_list.append(localfile.replace(".Z",""))
            Pot_compress_name_list = list(set(Pot_compress_name_list))
            
            for pot_compress_name in Pot_compress_name_list:
                if os.path.isfile(pot_compress_name):
                    Localfiles_lis.append(pot_compress_name)
    
    pool.close()
    return Localfiles_lis



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: motionpath.py,v 1.2 2006/01/27 07:51:02 mbaas Exp $

from cgkit.component import Component
from cgkit.slots import *
from math import *

# MotionPath
class MotionPath(Component):
    """Motion path.
    """
    
    def __init__(self,
                 name = "MotionPath",
                 curve = None,
                 begintime = 0.0,
                 endtime = 1.0,
                 loop = False,
                 follow = False,
                 bank = False,
                 bankamplitude = 0.1,
                 auto_insert = True):
        """Constructor.
        """
        Component.__init__(self, name=name, auto_insert=auto_insert)
        
        self.curve = curve

        self.begintime = begintime
        self.endtime = endtime
        self.loop = loop
        self.follow = follow
        self.bank = bank
        self.bankamplitude = bankamplitude

        self.transform_slot = ProceduralMat4Slot(self.computeTransform)
        self.time_slot = DoubleSlot()
        self.time_slot.addDependent(self.transform_slot)
        self.addSlot("transform", self.transform_slot)
        self.addSlot("time", self.time_slot)

    # computeTransform
    def computeTransform(self):
        """Procedural for the transform slot.
        """

        # Determine the native curve parameter t...
        time = self.time_slot.getValue()
        len = self.curve.length()
        s = (time-self.begintime)/(self.endtime-self.begintime)
        if self.loop:
            s %= 1.0
        s *= len
        if s<0:
            s = 0.0
        if s>len:
            s = len
        t = self.arcLenToCurveParam(s)

        # Evaluate the curve
        p, dp, d2p = self.curve.evalFrame(t)

        up = vec3(0,0,1)
        if self.follow:
            try:
                T = mat4().lookAt(p, p+dp, up)
            except:
                T = mat4().translation(p)
        else:
            T = mat4().translation(p)

        if self.bank:
            dp = dp.normalize()

            dt = 0.001
            dp_prev = self.curve.deriv(t-dt).normalize()
            d2p = (dp-dp_prev)/dt
            # Make the 2nd derivative orthogonal to the tangent...
            len = d2p.length()
            u = d2p.cross(dp)
            d2p = dp.cross(u)
            d2p = len*d2p.normalize()            
            if (dp.cross(up)*d2p<0):
                len = -len
            T = T*mat4().rotation(self.bankamplitude*len, vec3(0,0,1))
        return T

    # arcLenToCurveParam
    def arcLenToCurveParam(self, s, eps=0.0001, maxiter=100):
        """Determine the native curve parameter for a given arc length.
        """
        tmin, tmax = self.curve.paraminterval
        totallen = self.curve.arcLen(tmax)
        # Initial "guess"...
        t = tmin+(s/totallen)*(tmax-tmin)

        while maxiter>0:
            F = self.curve.arcLen(t)-s
            if abs(F)<=eps:
                return t

            dF = self.curve.deriv(t).length()
            if abs(dF)<1E-12:
                # TODO: do something
                print("MotionPath: ************ dF==0 !!!")
                dF = 1 
            t -= F/dF
            maxiter -= 1

        print(("MotionPath: Maximum number of iterations reached (s=%f)"%s))
        return t
        
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: joint.py,v 1.5 2005/08/28 19:42:43 mbaas Exp $

## \file joint.py
## Contains the Joint class.

from ._OpenGL.GL import *
from .cgtypes import *
from .component import *
from .worldobject import WorldObject
from .geomobject import GeomObject
from .spheregeom import SphereGeom
from .boundingbox import BoundingBox
from .euleradapter import EulerAdapter
from . import _core
from . import cmds
from .sl import *

# JointGeom
class JointGeom(GeomObject):
    """The geom object class for Joint objects.

    The geometry draws a wire frame sphere at the joint position
    and connections to all children joints (visualizing the bones). 
    """

    def __init__(self, joint, radius=0.05):
        """Constructor.

        \param joint (\c Joint) The joint that uses this geom
        \param radius (\c float) Radius of the wire sphere at the joint position
        """
        GeomObject.__init__(self)
        self.joint = joint
        self.radius = radius

        self.spheregeom = SphereGeom(radius=self.radius, segmentsu=8, segmentsv=4)

    def uniformCount(self):
        return 0

    def varyingCount(self):
        return 0

    def vertexCount(self):
        return 0

    def boundingBox(self):
        r = vec3(self.radius, self.radius, self.radius)
        return BoundingBox(-r, r)

    def drawGL(self):
        # Draw sphere
        glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT | GL_POLYGON_BIT)
        glDisable(GL_LIGHTING)
        glPushMatrix()

        # The joint is located at the pivot point
        p = self.joint.getOffsetTransform()[3]
        pivot = vec3(p.x, p.y, p.z)
        glTranslate(pivot.x, pivot.y, pivot.z)
        
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        self.spheregeom.drawGL()

        # Draw bone
        glBegin(GL_LINES)
        for child in self.joint.iterChilds():
            if not isinstance(child, Joint):
                continue
            p = child.pos-pivot
            try:
                b1 = p.ortho().normalize()
                b2 = p.cross(b1).normalize()
            except:
                continue
            b1 *= 0.75*self.radius
            b2 *= 0.75*self.radius
            
            glVertex3d(b1.x, b1.y, b1.z)
            glVertex3d(p.x, p.y, p.z)
            
            glVertex3d(-b1.x, -b1.y, -b1.z)
            glVertex3d(p.x, p.y, p.z)

            glVertex3d(b2.x, b2.y, b2.z)
            glVertex3d(p.x, p.y, p.z)

            glVertex3d(-b2.x, -b2.y, -b2.z)
            glVertex3d(p.x, p.y, p.z)
        glEnd()

        # Pivot coordinate system...
        P = self.joint.getOffsetTransform()
        r = 1.5*self.radius
        glBegin(GL_LINES)
        # X axis
        b = r*P[0]
        glColor3f(1,0,0)
        glVertex3f(0,0,0)
        glVertex3f(b.x,b.y,b.z)
        # Y axis
        b = r*P[1]
        glColor3f(0,1,0)
        glVertex3f(0,0,0)
        glVertex3f(b.x,b.y,b.z)
        # Z axis
        b = r*P[2]
        glColor3f(0,0,1)
        glVertex3f(0,0,0)
        glVertex3f(b.x,b.y,b.z)
        glEnd()

        glPopMatrix()
        glPopAttrib()


class Mult(Component):

    def __init__(self, name="Mult", auto_insert=True):
        Component.__init__(self, name=name, auto_insert=auto_insert)

        # Create the input slots
        self.op1_slot = Mat3Slot()
        self.op2_slot = Mat3Slot()
        # Create the output slot
        self.output_slot = ProceduralMat3Slot(self.computeOutput)

        # Add the slots to the component
        self.addSlot("op1", self.op1_slot)
        self.addSlot("op2", self.op2_slot)
        self.addSlot("output", self.output_slot)

        # Set up slot dependencies
        self.op1_slot.addDependent(self.output_slot)
        self.op2_slot.addDependent(self.output_slot)

    def computeOutput(self):
        return self.op1*self.op2

    # Create value attributes
    exec(slotPropertyCode("op1"))
    exec(slotPropertyCode("op2"))
    exec(slotPropertyCode("output"))

# Joint
class Joint(WorldObject):
    """Joint class.
    """

    def __init__(self,
                 name = "Joint",
                 radius = 0.05,
                 rotationorder = "xyz",
                 **params):
        
        WorldObject.__init__(self, name=name, **params)

        self.geom = JointGeom(self, radius=radius)

        self.euleradapter = EulerAdapter(order=rotationorder, outtype="mat3", auto_insert=False)
        self.nullrot = Mult(auto_insert=False)
        
#        self.euleradapter.output_slot.connect(self.rot_slot)
        self.nullrot.op1 = mat3(1)
        self.euleradapter.output_slot.connect(self.nullrot.op2_slot)
        self.nullrot.output_slot.connect(self.rot_slot)

        self.anglex_slot = self.euleradapter.anglex_slot
        self.angley_slot = self.euleradapter.angley_slot
        self.anglez_slot = self.euleradapter.anglez_slot

#    def setEuler(self, x,y,z):
#        m = mat3().fromEulerXYZ(radians(x), radians(y), radians(z))
#        self.rot = m

    ## protected:

    # angle properties...
    exec(slotPropertyCode("anglex"))
    exec(slotPropertyCode("angley"))
    exec(slotPropertyCode("anglez"))

    # freezePivot
    def freezePivot(self):
        """Make the current pivot coordinate system the default pose.

        After calling this method, the current rotation of the pivot
        coordinate system will define the default pose. This means,
        rotations are now defined around the local pivot axes.
        """
        self.nullrot.op1 = self.getOffsetTransform().getMat3()
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: scene.py,v 1.2 2005/04/14 07:58:55 mbaas Exp $

## \file scene.py
## Contains the Scene class.

from .cgtypes import vec3
from .Interfaces import ISceneItem
from . import timer
from . import protocols
from .boundingbox import BoundingBox
from . import joystick
from . import _core
from . import globalscene

# Scene
class Scene(object):
    """%Scene class."""
    
    def __init__(self):
        object.__init__(self)
        
        # Handedness ('r' or 'l')
        self._handedness = "r"
        # Normalized up direction
        self._up = vec3(0,0,1)
        # Unit   (1 unit = unitscale [Unit])
        self._unit = "m"
        # Unit scale
        self._unitscale = 1.0

        # Global options
        self._globals = {}
        
        self.items = []
#        self.items_by_name = {}

        self._worldroot = _core.WorldObject("WorldRoot")
        self._timer = timer.Timer(auto_insert=False)

        # Key: ID  Value:Joystick object
        self._joysticks = {}
        
        self.clear()

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def hasGlobal(self, name):
        return name in self._globals
            
    def getGlobal(self, name, default=None):
        return self._globals.get(name, default)

    def setGlobal(self, name, value):
        self._globals[name] = value

    # createDefaultItems
#    def createDefaultItems(self):
#        self._worldroot = WorldObject("root", auto_insert=False)

    # worldRoot
    def worldRoot(self):
        """Return the root object of the world.
        """
        return self._worldroot

    # walkWorld
    def walkWorld(self, root=None):
        """Walk the world tree and yield each object.

        This method can be used to iterator over the entire world tree or
        a subtree thereof. The argument root specifies the root of the
        tree which is to traverse (the root itself will not be returned).
        """
        if root==None:
            root = self._worldroot
        return self._walkWorld(root)

    # timer
    def timer(self):
        """Return the global timer object."""
        return self._timer

    # clear
    def clear(self):
        """Clear the entire scene."""

        # Remove all children of the world root...
        for obj in list(self._worldroot.iterChilds()):
            self._worldroot.removeChild(obj)
            
        self.items = [self._timer, self._worldroot]
#        self.items_by_name = {}

    # insert
    def insert(self, item):
        """Insert an item into the scene."""
#        protocols.adapt(item, ISceneItem)
#        if isinstance(item, worldobject.WorldObject):
        if isinstance(item, _core.WorldObject):
            self._worldroot.addChild(item)
        else:
            self.items.append(item)

    # item
    def item(self, name):
        """Return the item with the specified name.

        \param name (\c str) Item name
        \return Item or None if no item was found
        """
        for item in self.items:
            if name==item.name:
                return item
        return None

    # worldObject
    def worldObject(self, name):
        """Return the world object with the specified name.

        You can use '|' as a path separator.

        \param name (\c str) Object name
        \return WorldObject
        """
        obj = self._worldroot
        if len(name)>0:
            names = name.split("|")
        else:
            names = []
        for n in names:
            obj = obj.child(n)
        return obj

    # boundingBox
    def boundingBox(self):
        """Return the bounding box of the entire scene.

        \todo Use the world transform instead of the local transform
        """
        return self._worldroot.boundingBox()
#        return self._boundingBox(self._worldroot)
            
#    def _boundingBox(self, node):
#        res = BoundingBox()
#        for obj in node.iterChilds():
#            L = obj.localTransform()
#            g = obj.geom
#            if g!=None:
#                bb = g.boundingBox()
#                if not bb.isEmpty():
#                    res.addBoundingBox(bb.transform(L))

#            res.addBoundingBox(self._boundingBox(obj).transform(L))

#        return res

    # setJoystick
    def setJoystick(self, joystick):
        """Set a joystick object.
        """
        self._joysticks[joystick.id] = joystick

    # getJoystick
    def getJoystick(self, id):
        """Get a joystick object.

        A dummy joystick object is returned if there is no joystick
        with the specified id.
        """
        if id in self._joysticks:
            return self._joysticks[id]
        else:
            return joystick.Joystick(id=id, name="Dummy-Joystick")

    ## protected:


    def _walkWorld(self, obj):
        """Helper method for the walkWorld() method."""
        for child in obj.iterChilds():
            yield child
            for c in self._walkWorld(child):
                yield c

    
    # "handedness" property...
    
    def _getHandedness(self):
        """Return the handedness.

        This method is used for retrieving the \a handedness property.

        \return Handedness (\c str) which is either 'l' or 'r'.
        """
        return self._handedness

    def _setHandedness(self, h):
        """Set the handedness.

        This method is used for setting the \a handedness property.

        \param h (\c str) Handedness ('l' or 'r')
        """
        h = h.lower()
        if h!="l" and h!="r":
            raise ValueError("Handedness must be either 'l' or 'r'.")
        self._handedness = h

    handedness = property(_getHandedness, _setHandedness, None, "Handedness")

    # "up" property...
    
    def _getUp(self):
        """Return the up vector.

        This method is used for retrieving the \a up property.

        \return Up vector (\c vec3)
        """
        return self._up

    def _setUp(self, up):
        """Set the up vector.

        This method is used for setting the \a up property.

        \param up (\c vec3) Up vector
        """
        self._up = vec3(up).normalize()

    up = property(_getUp, _setUp, None, "Up vector")
    

######################################################################

# getScene
def getScene():
    """Return the global %scene instance."""
    return globalscene.getScene()




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: objmtl.py,v 1.3 2006/03/20 19:32:12 mbaas Exp $

from .cgtypes import *

# WavefrontReaderBase
class WavefrontReaderBase:
    """Wavefront OBJ/MTL reader base class.

    This is the base class for the OBJ and the MTL reader.
    """
    
    def __init__(self):
        """Constructor.
        """
        self.linenr = 0
        self.line = ""

    # read
    def read(self, f):
        """Read the content of a file.

        f is a file like object that can be used to read the content
        of the file.
        The file is read and for each keyword a handle_<keyword>() method
        is called with the data as argument (the number of arguments depends
        on the keyword). Each argument is of type str.
        A syntax error is generated if such a handler method is not available.
        The task of these handler methods is to preprocess the data, check
        for errors and invoke the final handler methods which are just called
        after the keyword.
        Before the file is read, the begin() method is called. At the end,
        the end() method is called.
        """

        self.linenr = 0
        self.begin()
        for self.line in f:
            self.linenr+=1
            line2 = self.line.strip()
            # Ignore empty lines and comments
            if line2=="" or line2[0] in ['#', '$', '!', '@']:
                continue

            a = line2.split()
            cmd = a[0]
            args = a[1:]
            handler = getattr(self, "handle_%s"%cmd, None)
            if handler!=None:
                handler(*args)
            else:
                self.handleUnknown(cmd, args)

        self.end()

    # begin
    def begin(self):
        """Begin reading a file.

        This method is called before the file is read.
        """
        pass

    # end
    def end(self):
        """End of reading.

        This method is called after the file was read.
        """
        pass

    # handleUnknown
    def handleUnknown(self, cmd, arglist):
        """Handle unknown keywords.

        cmd is the command keyword (the 1st argument in the current line)
        and arglist is the data (the remaining arguments).

        The default implementation raises a SyntaxError exception.
        """
        raise SyntaxError("Unknown statement in line %d: %s"%(self.linenr, self.line))


# MTLReader
class MTLReader(WavefrontReaderBase):
    """Wavefront MTL reader.

    This class can be used as base class to read a MTL file.
    """
    
    def __init__(self):
        """Constructor.
        """
        WavefrontReaderBase.__init__(self)

#    def handleUnknown(self, cmd, arglist):
#        pass
#        print "Unknown:",cmd,arglist

    # parseMapArgs
    def parseMapArgs(self, args):
        """Parse the arguments from a map_xyz line into options and names.

        args must be a list of strings containing the arguments (already
        split).
        Returns a 2-tuple (options, names) where options is a dictionary
        that contains the options that were present in args (key is
        the option string). names is a list of strings that were no option.
        """
        options = {}
        names = []
        while len(args)>0:
            a = args[0]
            # Is this an option?
            if a[0]=="-":
                if a=="-s":
                    val = tuple([float(x) for x in args[1:4]])
                    options[a] = val
                    args = args[4:]
                elif a=="-t":
                    val = tuple([float(x) for x in args[1:4]])
                    options[a] = val
                    args = args[4:]
                elif a=="-o":
                    val = tuple([float(x) for x in args[1:4]])
                    options[a] = val
                    args = args[4:]
                elif a=="-mm":
                    val = tuple([float(x) for x in args[1:3]])
                    options[a] = val
                    args = args[3:]
                elif a=="-bm":
                    options[a] = float(args[1])
                    args = args[2:]
                elif a=="-type":
                    options[a] = args[1]
                    args = args[2:]
                elif a=="-clamp" or a=="-blendu" or a=="-blendv":
                    options[a] = args[1].lower()=="on"
                    args = args[2:]
                else:
                    raise SyntaxError("Unknown map option in line %d: %s"%(self.linenr, a))
            else:  # no option
                names.append(a)
                args = args[1:]

        return options, names

    # Pre handler methods (they must be called "handle_<keyword>")

    def handle_newmtl(self, *name):
        """New material command.

        name is the name of the material (all names are concatenated).
        """
        self.newmtl(" ".join(name))

    def handle_illum(self, model):
        """Illumination model."""
        self.illum(int(model))

    def handle_Ns(self, shininess):
        """Shininess coefficient.
        """
        self.Ns(float(shininess))

    def handle_Ka(self, r, g, b):
        """Ambient color.
        """
        self.Ka(vec3(float(r), float(g), float(b)))

    def handle_Kd(self, r, g, b):
        """Diffuse color.
        """
        self.Kd(vec3(float(r), float(g), float(b)))

    def handle_Ks(self, r, g, b):
        """Specular color.
        """
        self.Ks(vec3(float(r), float(g), float(b)))

    def handle_Ke(self, r, g, b):
        """Emissive color.
        """
        self.Ke(vec3(float(r), float(g), float(b)))

    def handle_Tr(self, alpha):
        """Transparency."""
        self.Tr(float(alpha))

    def handle_d(self, alpha):
        """Dissolve factor (transparency)."""
        self.d(float(alpha))

    def handle_Tf(self, *args):
        """Transparency (as color)."""
        if len(args)==1:
            v = vec3(float(args[0]))
        else:
            r,g,b = args
            v = vec3(float(r), float(g), float(b))
        
        self.Tf(v)

    def handle_Ni(self, ref):
        """Refraction index."""
        self.Ni(float(ref))

    def handle_sharpness(self, v):
        """Sharpness value."""
        self.sharpness(float(v))

    def handle_map_Ka(self, *mapargs):
        """Ambient texture file."""
        opts, args = self.parseMapArgs(mapargs)
        if len(args)>0:
            self.map_Ka(args[0], opts)

    def handle_map_Kd(self, *mapargs):
        """Diffuse texture file."""
        opts, args = self.parseMapArgs(mapargs)
        if len(args)>0:
            self.map_Kd(args[0], opts)

    def handle_map_Ks(self, *mapargs):
        """Specular texture file."""
        opts, args = self.parseMapArgs(mapargs)
        if len(args)>0:
            self.map_Ks(args[0], opts)

    def handle_map_Ke(self, *mapargs):
        """Emission texture file."""
        opts, args = self.parseMapArgs(mapargs)
        if len(args)>0:
            self.map_Ke(args[0], opts)

    def handle_map_Ns(self, *mapargs):
        """Shininess texture file."""
        opts, args = self.parseMapArgs(mapargs)
        if len(args)>0:
            self.map_Ns(args[0], opts)

    def handle_map_d(self, *mapargs):
        """Opacity texture file."""
        opts, args = self.parseMapArgs(mapargs)
        if len(args)>0:
            self.map_d(args[0], opts)

    def handle_map_Bump(self, *mapargs):
        """Bump map texture file."""
        opts, args = self.parseMapArgs(mapargs)
        if len(args)>0:
            self.map_Bump(args[0], opts)

    def handle_refl(self, *mapargs):
        """Reflection map."""
        opts, args = self.parseMapArgs(mapargs)
        if len(args)>0:
            self.refl(args[0], opts)

    # Handler methods

    def newmtl(self, name):
        pass

    def illum(self, model):
        """Illumination model."""
        pass

    def Ns(self, shininess):
        """Shininess."""
        pass

    def Ka(self, c):
        """Ambient color.

        c is a vec3 containing the color.
        """
        pass

    def Kd(self, c):
        """Diffuse color.

        c is a vec3 containing the color.
        """
        pass

    def Ks(self, c):
        """Specular color.

        c is a vec3 containing the color.
        """
        pass

    def Ke(self, c):
        """Emissive color.

        c is a vec3 containing the color.
        """
        pass

    def Tr(self, alpha):
        """Transparency."""
        pass

    def d(self, alpha):
        """Dissolve factor (transparency)."""
        pass

    def Tf(self, c):
        """Transparency.

        c is a vec3 containing a color.
        """
        pass

    def Ni(self, ref):
        """Refraction index."""
        pass

    def sharpness(self, v):
        """Sharpness value."""
        pass

    def map_Ka(self, mapname, options):
        pass

    def map_Kd(self, mapname, options):
        pass

    def map_Ks(self, mapname, options):
        pass

    def map_Ke(self, mapname, options):
        pass

    def map_Ns(self, mapname, options):
        pass

    def map_d(self, mapname, options):
        pass

    def map_Bump(self, mapname, options):
        pass

    def refl(self, mapname, options):
        pass


# OBJReader
class OBJReader(WavefrontReaderBase):
    """Wavefront OBJ reader.

    This class can be used as base class to read an OBJ file.
    """
    
    def __init__(self):
        """Constructor.
        """
        WavefrontReaderBase.__init__(self)
        self.v_count = 0
        self.vp_count = 0
        self.vt_count = 0
        self.vn_count = 0

    # read
    def read(self, f):
        self.v_count = 0
        self.vp_count = 0
        self.vt_count = 0
        self.vn_count = 0
        WavefrontReaderBase.read(self, f)

    # Pre handler methods (they must be called "handle_<keyword>")

    def handle_mtllib(self, *files):
        """Material library command.

        files contains the filenames.
        """
        if len(files)==0:
            raise SyntaxError("No material library given in line %d: %s"%(self.linenr, self.line))
        self.mtllib(*files)

    def handle_usemtl(self, *name):
        """Material name.

        (all names are concatenated)
        """
        self.usemtl(" ".join(name))

    def handle_g(self, *groups):
        """Group command.

        groups are the group names that the following geometry belongs to.
        """
        if len(groups)==0:
            groups = ("default",)
#            raise SyntaxError("No group name given in line %d: %s"%(self.linenr, self.line))
        self.g(*groups)

    def handle_s(self, group_number):
        """Smoothing group command.

        group_number is a string that contains the smoothing group number
        or "off".
        """
        group_number = group_number.lower()
        if group_number=="off":
            group_number = 0
        try:
            group_number = int(group_number)
        except:
            raise SyntaxError("Invalid smoothing group number in line %d: %s"%(self.linenr, self.line))
        self.s(group_number)

    def handle_v(self, x, y, z, w=1):
        """Vertex definition.
        """
        self.v_count += 1
        self.v(vec4(float(x), float(y), float(z), float(w)))

    def handle_vp(self, u, v=0, w=1):
        """Vertex in parameter space.
        """
        self.vp_count += 1
        self.vp(vec3(float(u), float(v), float(w)))

    def handle_vn(self, x, y, z):
        """Normal.
        """
        self.vn_count += 1
        self.vn(vec3(float(x), float(y), float(z)))

    def handle_vt(self, u, v=0, w=0):
        """Texture vertex.
        """
        self.vt_count += 1
        self.vt(vec3(float(u), float(v), float(w)))

    def handle_p(self, *verts):
        """Points."""
        
        if len(verts)==0:
            raise SyntaxError("At least 1 vertex required in line %d: %s"%(self.linenr, self.line))

        vlist = []
        for s in verts:
            vert = int(s)
            if vert<0:
                vert = self.v_count+vert+1
            if vert==0:
                raise ValueError("0-index in line %d: %s"%(self.linenr, self.line))
            vlist.append(vert)
            
        self.p(*vlist)     

    def handle_l(self, *verts):
        """Line."""

        if len(verts)<2:
            raise SyntaxError("At least 2 vertices required in line %d: %s"%(self.linenr, self.line))

        vlist = []
        for s in verts:
            a = s.split("/")
            if len(a)==0 or len(a)>2:
                raise SyntaxError("Syntax error in line %d: %s"%(self.linenr, self.line))
            vert = int(a[0])
            if vert<0:
                vert = self.v_count+vert+1
            tvert = None
            if len(a)>1:
                if a[1]!="":
                    tvert = int(a[1])
                    if tvert<0:
                        tvert = self.vt_count+tvert+1
            if vert==0 or tvert==0:
                raise ValueError("0-index in line %d: %s"%(self.linenr, self.line))
            vlist.append((vert, tvert))
        self.l(*vlist)

    def handle_f(self, *verts):
        """Polygonal face.
        """
        if len(verts)<3:
            raise SyntaxError("At least 3 vertices required in line %d: %s"%(self.linenr, self.line))

        vlist = []
        for s in verts:
            a = s.split("/")
            if len(a)==0 or len(a)>3:
                raise SyntaxError("Syntax error in line %d: %s"%(self.linenr, self.line))
            vert = int(a[0])
            if vert<0:
                vert = self.v_count+vert+1
            tvert = None
            normal = None
            if len(a)>1:
                if a[1]!="":
                    tvert = int(a[1])
                    if tvert<0:
                        tvert = self.vt_count+tvert+1
                if len(a)>2 and a[2]!="":
                    normal = int(a[2])
                    if normal<0:
                        normal = self.vn_count+normal+1
            if vert==0 or tvert==0 or normal==0:
                raise ValueError("0-index in line %d: %s"%(self.linenr, self.line))
            vlist.append((vert, tvert, normal))
        self.f(*vlist)

    def handle_o(self, name):
        """Object name.
        """
        self.o(name)

    def handle_bevel(self, on_off):
        """Bevel interpolation on/off.
        """
        self.bevel(on_off)

    def handle_c_interp(self, on_off):
        """Color interpolation on/off.
        """
        self.c_interp(on_off)

    def handle_d_interp(self, on_off):
        """Dissolve interpolation on/off.
        """
        self.d_interp(on_off)

    def handle_lod(self, level):
        """Level of Detail.
        """
        self.lod(int(level))

    def handle_shadow_obj(self, filename):
        """Shadow object.
        """
        self.shadow_obj(filename)

    def handle_trace_obj(self, filename):
        """Shadow object.
        """
        self.trace_obj(filename)

    # Handler methods

    def call(self, filename, *args):
        pass

    def csh(self, cmd):
        pass

    def mtllib(self, *files):
        """Specification of material libraries.

        files is a sequence of file names that contain material definitions.
        """
        pass

    def usemtl(self, name):
        """Material name.

        name is a string containing the name of the material to use for
        the following elements.
        """
        pass

    def g(self, *groups):
        """Grouping statement.

        groups is a sequence of group names that the following geometry
        belongs to.
        """
        pass

    def s(self, group_number):
        """Smoothing group.

        group_number is an integer containing the smoothing group number.
        Smoothing groups should be turned off if the group number is 0.
        """
        pass

    def v(self, vert):
        """Geometric vertex.

        vert is always a vec4."""
        pass

    def vp(self, vert):
        """A point in parameter space.

        This vertex is used for free-form curves or surfaces.
        vert is always a vec3.
        """
        pass

    def vn(self, normal):
        """Normal vector.
        
        normal is a vec3.
        """
        pass

    def vt(self, tvert):
        """Texture vertex.
        
        tvert is always a vec3.
        """
        pass

    def p(self, *verts):
        """Points.

        verts is a list of vertex indices. The indices are always >0
        (negative indices are automatically converted to their
        corresponding positive indices).
        All indices are 1-based. If an index in the file was 0, an exception
        was already thrown.
        """
        print(("POINT",verts))

    def l(self, *verts):
        """Line.

        verts contains 2-tuples (vert, tvert) which contains the indices
        to the vertexa and the texture vertex. tvert may be None.
        The indices are always >0 (negative indices are automatically
        converted to their corresponding positive indices).
        All indices are 1-based. If an index in the file was 0, an exception
        was already thrown.
        """
        pass

    def f(self, *verts):
        """Polygonal face.
        
        verts contains 3-tuples (vert, tvert, normal) which contains
        the indices to the vertex, the texture vertex and the normal.
        tvert and normal may be None, otherwise the values are always >0
        (negative indices are automatically converted to their corresponding
        positive indices).
        All indices are 1-based. If an index in the file was 0, an exception
        was already thrown.
        """
        pass

    def o(self, name):
        """Optional object name.

        name is a string containing the specified name for the elements
        following this statement.
        """
        pass

    def bevel(self, on_off):
        pass

    def c_interp(self, on_off):
        pass

    def d_interp(self, on_off):
        pass

    def lod(self, level):
        """
        level is an integer.
        """
        pass

    def shadow_obj(self, filename):
        pass

    def trace_obj(self, filename):
        pass

